/*
 * This class is based on Krislet's Brain and provides offline training/validation services to the KrisletScenes client.
 * It can be used as a testbed for statistically evaluating the performance of DistanceCalculation objects.
 * 
 * Created on Jan 26, 2005
 *
 */
package rcscene;

import java.util.*;
import java.io.*;

import sceneInfo.*;
import sceneMatch.nearest_neighbor_matching.*;
import sceneMatch.sceneMatchAlgorithm;

/**
 * @author Kevin Lam
 *
 */
class Validator
{
	
	private SceneLib storedScenes;
	private SceneLib validateScenes;
	//private SceneLib sceneLib = null;
	
	int numRows = 0;
	int numCols = 0;

	int validationIterations = 0;
	
	SceneStats gameStats;
	
	public Validator(String scenelib, String comparelib, int sceneSel, 
					int distCal, int numBest, float[] objWeights, float[] actionWeights,
					int testIterations)
	{
		m_side = 'l';	// arbitrary, shouldn't matter

		m_distCal = distCal;
		m_sceneSel = sceneSel;
		m_numBest = numBest;
		m_objWeights = objWeights;
		m_actionWeights = actionWeights;

		// initialize scene database
		System.out.println("Loading stored scene file " + scenelib);
		storedScenes = loadSceneFile(scenelib);
		
		int srows = numRows;
		int scols = numCols;
		
		validationIterations = testIterations;
		
		if(comparelib.compareToIgnoreCase(scenelib) == 0)
		{
			System.out.println("Using the same scene file for target");
			validateScenes = storedScenes;
		}
		else
		{
			System.out.println("Loading target scene file " + comparelib);
			validateScenes = loadSceneFile(comparelib);
			System.out.println("Read " + validateScenes.size() + " scenes loaded for validation.");

		} 
   		
   		// double check scene files match in dimension
		if((srows != numRows)||(scols != numCols))
		{
			System.err.println("WARNING - scene files differ in cell dimension");
		}
		
		
   		//create a new stats object to keep track of the 
   		//number of scenes processed and time taken to process these
   		//scenes
   		gameStats = new SceneStats();
   		
	}

	// for a future version maybe try to encapsulate all these into a single utility class that other clients just hook into...
	// ... an item for refactoring later
	public SceneLib loadSceneFile(String fname)
		{
			SceneLib loadedlist = null;
			System.out.println("Loading scenes from file " + fname);
			try
			{
					FileInputStream fileIn = new FileInputStream(fname);
					ObjectInputStream ois = new ObjectInputStream(fileIn);
					
					numRows = ((Integer) ois.readObject()).intValue();
					numCols = ((Integer) ois.readObject()).intValue();
					
					//storedScenes = (Vector) ois.readObject();
					//target = (ArrayList) ois.readObject();
					//validateScenes = (ArrayList) ois.readObject();
					loadedlist = (SceneLib) ois.readObject();
					
					ois.close();
					fileIn.close();
			}
			catch (FileNotFoundException e)
			{
				e.printStackTrace();
			}
			catch (IOException e)
			{
				e.printStackTrace();
			}
			catch (ClassNotFoundException e)
			{
				e.printStackTrace();
			}
		
			//System.out.println("Read " + target.size() + " scenes, table size (" + numRows + "," + numCols + ")");
			
			return loadedlist;
			
		}
		
	//  1. Get next Scene from target list
	//  3. Instantiate a DistanceCalculation object
	//  4. Pick out the best scene from the list
	//  5. Compare chosen action with stored action and get statistics
	//
	public void runValidation()
	{
		long startTime = 0, endTime = 0, totalTime = 0;
		float totalMatches = 0, totalCompares = 0;
		float origKick = 0, correctKick = 0;
		int totalScenes = 0;
		
		System.out.println("Beginning validation of " + validateScenes.size() + " scenes");
		
		
		DistanceCalculation distanceCalc;			// plug-and-play distance calculation object
				
		// *********************************************
		// choose scene selection algorithm, parameters
		// *********************************************
		switch(m_distCal) {
		  case 0: {
			  distanceCalc = new NearestNeighborCartesianObjects(m_numBest, m_side);
			  //distanceCalc = new NearestNeighborBallGoalDistance(m_numBest, m_side);
			  break;
			}
		  case 1: {
			distanceCalc = new NearestNeighborCartesianCellObjects(m_numBest, m_side);
			break;
		  }
		  case 2: {
			distanceCalc = new NearestNeighborCellBallGoalDistance(m_numBest, m_side);
			break;
		  }
		  case 3: {
			distanceCalc = new RandomDistance(m_numBest, m_side);
			break;
		  }
		  case 4: { // bipartite matching
			distanceCalc = new sceneMatchAlgorithm();
			break;
		  }
		  default: {
			distanceCalc = new NearestNeighborCartesianCellObjects(m_numBest, m_side);
			break;
		  }
		}
		
		// load weights
		distanceCalc.setObjectWeights(m_objWeights);
		
		
		int tempcount = 0;
		String currentStep = "";
		int sceneCount = 0;
		
		try {
			
		// iterate through scenes
		Iterator it = validateScenes.getSceneList().iterator();
		while( it.hasNext() )
		{
			sceneCount++;
			currentStep = "Starting While Loop";
			
			//if (tempcount++ == 1000) break;
			
			currentStep = "Getting next scene from validation set";
			//get the next scene from the validation set
			Scene thisScene = (Scene) it.next();
			
			/////////////////////////////////////////////////////////////////
			//get the ball distance... only other change needed for the 
			//new scene lib to work correctly.. 
			/*VisualInfo stuff = thisScene.getVision();
			if (stuff.getBallList() == null) {
				stuff.setSceneDist(0.0f);
			} else {
				if (!stuff.getBallList().isEmpty()) {
					BallInfo tmpinfo = (BallInfo)stuff.getBallList().get(0);
					stuff.setSceneDist(tmpinfo.m_distance);
				} else {
					stuff.setSceneDist(0.0f);
				}
			}*/
			/////////////////////////////////////////////////////////////////
		
			VisualInfo newValScene = thisScene.getVision();
			VisualInfo stuff = new VisualInfo(newValScene.m_message, "ValidateTeam", false);
			stuff.setNodeOffset((short)30);
			stuff.parse();
			if (stuff.getBallInfo() == null) {
				stuff.setSceneDist(250.0f);
			} else {
				BallInfo tmpbll = stuff.getBallInfo();
				if (tmpbll.m_distance > 0.0f) {
					stuff.setSceneDist(tmpbll.m_distance);
				} else {
					stuff.setSceneDist(0.0f);
				}
				
			}
			
			
			totalScenes++;
			
			currentStep = "Getting cycle start time";
			
			// record start time
			startTime = System.currentTimeMillis();
			
			
			currentStep = "Calling matching algorithm";
			////////////////////////////////////////////////////////////////////
			//get the closest matching scene from the close set
						
			Scene[] bestScenes = distanceCalc.findClosest(thisScene, storedScenes.getSceneList());

			
			// result is a set of candidate scenes stored in bestScenes; now, choose the best one to imitate
			//explain(bestScenes, thisScene, distanceCalc);
			
			// *********************************************
			// choose scene selection algorithm, parameters
			// *********************************************
			
			Scene matched;
			
			switch(m_sceneSel) {
			  case 1: {
				matched = SceneSelection.ChooseRandom(bestScenes, m_numBest);
				break;
			  }
			  case 2: {
				matched = SceneSelection.VoteWeighted(bestScenes, bestScenes.length, m_actionWeights);
				break;
			  }
			  case 3: {
				matched = SceneSelection.VoteWeightedWithRandom(bestScenes, bestScenes.length, m_actionWeights, 20);
				break;
			  }
			  default: {	// 0
				matched = SceneSelection.ChooseFirstValid(bestScenes);
				break;
			  }
			}
			
			
			currentStep = "Getting cycle end time";
			
			// record end time
			endTime = System.currentTimeMillis();
			//System.out.println("Time taken: " + (endTime - startTime));
			totalTime += (endTime - startTime);

			// now compare the actions in thisScene with the actions in matched; track whether we got them right or not
			//System.out.println("Scene " + thisScene.getIdentString() + "/" + thisScene.getActions() +
			  //                 " matches best with scene " + matched.getIdentString() + "/" + matched.getActions());
			               
			currentStep = "Getting matched scene actions";
			//record this cycles times
			//gameStats.addSceneInfo(distanceCalc.getSceneCount(), (endTime - startTime), matched.getIdentString(), bestScenes.length);
			ArrayList matchedActions = matched.getActions();
			String matchedActionString = "";
			if (matchedActions.size() == 1) {
				matchedActionString += ((Action)matchedActions.get(0)).toString();
			} else if (matchedActions.size() == 2) {
				matchedActionString += ((Action)matchedActions.get(0)).toString();
				matchedActionString += ((Action)matchedActions.get(1)).toString();
			} else if (matchedActions.size() > 2) {
				matchedActionString += ((Action)matchedActions.get(0)).toString();
				matchedActionString += ((Action)matchedActions.get(1)).toString();
				matchedActionString += "...";
			} 
			
			currentStep = "Saving matched scene information";
			gameStats.addSceneInfo(storedScenes.size(), (endTime - startTime), matched.getIdentString(), storedScenes.size(), matchedActionString);
			
			currentStep = "Getting action from test scene";
			ArrayList knownActions = thisScene.getActions();
			Action testAction = (Action)knownActions.get(0);
			
			switch (testAction.getAction()) {
			case Action.ACTION_NONE:
				break;
			case Action.ACTION_DASH:
				gameStats.addCorrectDASH();
				break;
			case Action.ACTION_KICK:
				gameStats.addCorrectKICK();
				break;
			case Action.ACTION_TURNNECK:
				gameStats.addCorrectTURNHEAD();
				break;
			case Action.ACTION_TURN:
				gameStats.addCorrectTURN();
				break;
			case Action.ACTION_CATCH:
				break;
			case Action.ACTION_MOVE:
				break;
			}
			
			currentStep = "Getting matched action";
			knownActions = matched.getActions();
			Action matchedAction = (Action)knownActions.get(0);
			
			if (matchedAction.getAction() == testAction.getAction())
			{
				gameStats.addSceneHit(matched.getIdentString());
				
				//add checks for more actions and record them
				//to the scenestats object
				switch (matchedAction.getAction()) {
					case Action.ACTION_NONE:
						break;
					case Action.ACTION_DASH:
						gameStats.addMatchedDASH();
						break;
					case Action.ACTION_KICK:
						gameStats.addMatchedKICK();
						break;
					case Action.ACTION_TURNNECK:
						gameStats.addMatchedTURNHEAD();
						break;
					case Action.ACTION_TURN:
						gameStats.addMatchedTURN();
						break;
					case Action.ACTION_CATCH:
						break;
					case Action.ACTION_MOVE:
						break;
				}
				
				gameStats.addHit();
				
				// we have a matching action, so count it, and remove the items from the lists so they won't be counted twice
				totalMatches++;
			} else {
				gameStats.addSceneMiss(matched.getIdentString());
				
				//did not match, but will record what was selected in the
				//event of a miss
				switch (matchedAction.getAction()) {
					case Action.ACTION_NONE:
						break;
					case Action.ACTION_DASH:
						gameStats.addSelectedDASH();
						break;
					case Action.ACTION_KICK:
						gameStats.addSelectedKICK();
						break;
					case Action.ACTION_TURNNECK:
						gameStats.addSelectedTURNHEAD();
						break;
					case Action.ACTION_TURN:
						gameStats.addSelectedTURN();
						break;
					case Action.ACTION_CATCH:
						break;
					case Action.ACTION_MOVE:
						break;
				}			
				
				gameStats.addMiss();
				
			}
		
			
			
			/*
			 * will compare on the first action of the test scene
			 * and the matched scene. looping like this introduces redundancies and
			 * errors into the reporting, plus is bloats the numbers, because
			 * many more 'actions' are reached than scenes. ideally should be
			 * one to one.
			 * 
			currentStep = "Getting Vector of Actions from validation scene";
			Vector knownActions = new Vector(thisScene.getActions());
			Iterator knownList = knownActions.iterator();
			while (knownList.hasNext())
			{			
				currentStep = "Saving stats on validation action";
				Action compareThis = (Action) knownList.next();
				
				//if(compareThis.getAction() == Action.ACTION_KICK)
				//	origKick++;
				switch (compareThis.getAction()) {
					case Action.ACTION_NONE:
						break;
					case Action.ACTION_DASH:
						gameStats.addCorrectDASH();
						break;
					case Action.ACTION_KICK:
						gameStats.addCorrectKICK();
						break;
					case Action.ACTION_TURNNECK:
						gameStats.addCorrectTURNHEAD();
						break;
					case Action.ACTION_TURN:
						gameStats.addCorrectTURN();
						break;
					case Action.ACTION_CATCH:
						break;
					case Action.ACTION_MOVE:
						break;
				}

				currentStep = "Getting Vector of Actions from matched scene";
				Vector actions = new Vector(matched.getActions());
				Iterator matchList = actions.iterator();
				while (matchList.hasNext())
				{
					Action doThis = (Action) matchList.next();
					
					if (doThis.getAction() == compareThis.getAction())
					{
						gameStats.addSceneHit(matched.getIdentString());
						
						//if(doThis.getAction() == Action.ACTION_KICK)
						//	correctKick++;

						//add checks for more actions and record them
						//to the scenestats object
						switch (doThis.getAction()) {
							case Action.ACTION_NONE:
								break;
							case Action.ACTION_DASH:
								gameStats.addMatchedDASH();
								break;
							case Action.ACTION_KICK:
								gameStats.addMatchedKICK();
								break;
							case Action.ACTION_TURNNECK:
								gameStats.addMatchedTURNHEAD();
								break;
							case Action.ACTION_TURN:
								gameStats.addMatchedTURN();
								break;
							case Action.ACTION_CATCH:
								break;
							case Action.ACTION_MOVE:
								break;
						}
						
						gameStats.addHit();
						
						// we have a matching action, so count it, and remove the items from the lists so they won't be counted twice
						totalMatches++;
						knownList.remove();
						matchList.remove();
						break;
					} else {
						gameStats.addSceneMiss(matched.getIdentString());
						
						//did not match, but will record what was selected in the
						//event of a miss
						switch (doThis.getAction()) {
							case Action.ACTION_NONE:
								break;
							case Action.ACTION_DASH:
								gameStats.addSelectedDASH();
								break;
							case Action.ACTION_KICK:
								gameStats.addSelectedKICK();
								break;
							case Action.ACTION_TURNNECK:
								gameStats.addSelectedTURNHEAD();
								break;
							case Action.ACTION_TURN:
								gameStats.addSelectedTURN();
								break;
							case Action.ACTION_CATCH:
								break;
							case Action.ACTION_MOVE:
								break;
						}			
						
						gameStats.addMiss();
						
					}
				}
				totalCompares++;
			}
			*/
			
			if ((sceneCount == validationIterations) && (validationIterations > 0)) {
				break;
			}
			
		}
		
		} catch (StackOverflowError e) {
			System.out.println(e);
			e.printStackTrace();
			System.out.println(System.err.toString());
			System.out.println("Current loop interation: " + sceneCount);
			System.out.println("Current step: " + currentStep);
			
		}	
		
		
		// output statistics
		System.out.println("\n\nValidation complete.\n");
		
		//System.out.println("Compared " + totalScenes + " scenes against database of " + storedScenes.size() + " scenes");
		//System.out.println("Total action comparisons: " + (int)totalCompares);
		//System.out.println("Total correct actions: " + (int)totalMatches + "(" + (totalMatches / totalCompares)*100 + "%)");
		//System.out.println("Total KICK actions: " + (int)origKick);
		//System.out.println("Total correct KICKs: " + (int)correctKick + "(" + (correctKick / origKick)*100 + "%)");
		//System.out.println("Average time taken: " + totalTime / totalScenes + " mill. per scene");
		
		
		//when leaves the run, should be the end of the game - 
		//display the stats for this game
		gameStats.calcAvgs();
		System.out.println(gameStats.getStats());
		
		return;
	}


/**
	 * @param bestScenes
	 */
	private void explain(Scene[] bestScenes, Scene target, DistanceCalculation distcalc)
	{
		System.out.println("Best " + bestScenes.length + " matches to scene " + target.getIdentString());
		for(int i = 0; i < bestScenes.length; i++)
		{
			Scene x = bestScenes[i];
			System.out.println("   scene " + x.getIdentString() + ", distance " + distcalc.distanceBetween(x, target));		
		}
		System.out.println("Distance to self is " + distcalc.distanceBetween(target, target));				
		
	}

	//===========================================================================
// Private members
	private char					m_side;
	private	int						m_distCal;
	private int						m_sceneSel;
	private int						m_numBest;
	private float[]					m_objWeights;
	private float[]					m_actionWeights;
	
}
